kinetics_templates_action_clip = [
    f"a photo of action {{}}",
    f"a picture of action {{}}",
    f"Human action of {{}}",
    f"{{}}, an action",
    f"{{}} this is an action",
    f"{{}}, a video of action",
    f"Playing action of {{}}",
    f"{{}}",
    f"Playing a kind of action, {{}}",
    f"Doing a kind of action, {{}}",
    f"Look, the human is {{}}",
    f"Can you recognize the action of {{}}?",
    f"Video classification of {{}}",
    f"A video of {{}}",
    f"The man is {{}}",
    f"The woman is {{}}",
]

kinetics_templates = [
    "A photo of action {}.",
    "A video of action {}.",
    "He or she is {}.",
    "A person is doing {}.",
    "Look, the human is {}.",
    "Human action of {}.",
    "Playing action of {}.",
    "Video classification of {}.",
    "Doing a kind of action, {}.",
    "Playing a kind of action, {}.",
    "Can you recognize the action of {}?",
    "{}, an action.",
    "{} this is an action.",
    "{}, a video of action.",
    "An action of {} is in the video.",
    "There is a person doing {} in the video.",
    "A photo of a person doing {}.",
    "A photo of a person performing {}.",
    "A photo of a person practicing {}.",
    "A video of a person doing {}.",
    "A video of a person performing {}.",
    "A video of a person practicing {}.",
    "A example of a person doing {}.",
    "A example of a person performing {}.",
    "A example of a person practicing {}.",
    "A demonstration of a person doing {}.",
    "A demonstration of a person performing {}.",
    "A demonstration of a person practicing {}.",
]

imagenet_templates = [
    "a bad photo of a {}.",
    "a photo of many {}.",
    "a sculpture of a {}.",
    "a photo of the hard to see {}.",
    "a low resolution photo of the {}.",
    "a rendering of a {}.",
    "graffiti of a {}.",
    "a bad photo of the {}.",
    "a cropped photo of the {}.",
    "a tattoo of a {}.",
    "the embroidered {}.",
    "a photo of a hard to see {}.",
    "a bright photo of a {}.",
    "a photo of a clean {}.",
    "a photo of a dirty {}.",
    "a dark photo of the {}.",
    "a drawing of a {}.",
    "a photo of my {}.",
    "the plastic {}.",
    "a photo of the cool {}.",
    "a close-up photo of a {}.",
    "a black and white photo of the {}.",
    "a painting of the {}.",
    "a painting of a {}.",
    "a pixelated photo of the {}.",
    "a sculpture of the {}.",
    "a bright photo of the {}.",
    "a cropped photo of a {}.",
    "a plastic {}.",
    "a photo of the dirty {}.",
    "a jpeg corrupted photo of a {}.",
    "a blurry photo of the {}.",
    "a photo of the {}.",
    "a good photo of the {}.",
    "a rendering of the {}.",
    "a {} in a video game.",
    "a photo of one {}.",
    "a doodle of a {}.",
    "a close-up photo of the {}.",
    "a photo of a {}.",
    "the origami {}.",
    "the {} in a video game.",
    "a sketch of a {}.",
    "a doodle of the {}.",
    "a origami {}.",
    "a low resolution photo of a {}.",
    "the toy {}.",
    "a rendition of the {}.",
    "a photo of the clean {}.",
    "a photo of a large {}.",
    "a rendition of a {}.",
    "a photo of a nice {}.",
    "a photo of a weird {}.",
    "a blurry photo of a {}.",
    "a cartoon {}.",
    "art of a {}.",
    "a sketch of the {}.",
    "a embroidered {}.",
    "a pixelated photo of a {}.",
    "itap of the {}.",
    "a jpeg corrupted photo of the {}.",
    "a good photo of a {}.",
    "a plushie {}.",
    "a photo of the nice {}.",
    "a photo of the small {}.",
    "a photo of the weird {}.",
    "the cartoon {}.",
    "art of the {}.",
    "a drawing of the {}.",
    "a photo of the large {}.",
    "a black and white photo of a {}.",
    "the plushie {}.",
    "a dark photo of a {}.",
    "itap of a {}.",
    "graffiti of the {}.",
    "a toy {}.",
    "itap of my {}.",
    "a photo of a cool {}.",
    "a photo of a small {}.",
    "a tattoo of the {}.",
]
